@@ -11,13 +11,14 @@
 
 #include <string.h>
 
+#include <glib/gi18n-lib.h>
 #include <gio/gnetworking.h>
 
 #include "soup-socket.h"
+#include "soup-socket-private.h"
 #include "soup.h"
 #include "soup-filter-input-stream.h"
 #include "soup-io-stream.h"
-#include "soup-misc-private.h"
 
 /**
  * SECTION:soup-socket
@@ -29,7 +30,11 @@
  * soup_socket_get_remote_address()) may be useful to applications.
  **/
 
-G_DEFINE_TYPE (SoupSocket, soup_socket, G_TYPE_OBJECT)
+static void soup_socket_initable_interface_init (GInitableIface *initable_interface);
+
+G_DEFINE_TYPE_WITH_CODE (SoupSocket, soup_socket, G_TYPE_OBJECT,
+			 G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
+						soup_socket_initable_interface_init))
 
 enum {
 	READABLE,
@@ -45,9 +50,13 @@ static guint signals[LAST_SIGNAL] = { 0 };
 enum {
 	PROP_0,
 
+	PROP_FD,
+	PROP_GSOCKET,
+	PROP_IOSTREAM,
 	PROP_LOCAL_ADDRESS,
 	PROP_REMOTE_ADDRESS,
 	PROP_NON_BLOCKING,
+	PROP_IPV6_ONLY,
 	PROP_IS_SERVER,
 	PROP_SSL_CREDENTIALS,
 	PROP_SSL_STRICT,
@@ -56,10 +65,9 @@ enum {
 	PROP_USE_THREAD_CONTEXT,
 	PROP_TIMEOUT,
 	PROP_TRUSTED_CERTIFICATE,
-	PROP_CLEAN_DISPOSE,
 	PROP_TLS_CERTIFICATE,
 	PROP_TLS_ERRORS,
-	PROP_PROXY_RESOLVER,
+	PROP_SOCKET_PROPERTIES,
 
 	LAST_PROP
 };
@@ -71,9 +79,11 @@ typedef struct {
 	GInputStream *istream;
 	GOutputStream *ostream;
 	GTlsCertificateFlags tls_errors;
+	GTlsInteraction *tls_interaction;
 	GProxyResolver *proxy_resolver;
 
 	guint non_blocking:1;
+	guint ipv6_only:1;
 	guint is_server:1;
 	guint ssl:1;
 	guint ssl_strict:1;
@@ -90,12 +100,15 @@ typedef struct {
 	guint timeout;
 
 	GCancellable *connect_cancel;
+	int fd;
 } SoupSocketPrivate;
 #define SOUP_SOCKET_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SOUP_TYPE_SOCKET, SoupSocketPrivate))
 
 static void soup_socket_peer_certificate_changed (GObject *conn,
 						  GParamSpec *pspec,
 						  gpointer user_data);
+static void finish_socket_setup (SoupSocket *sock);
+static void finish_listener_setup (SoupSocket *sock);
 
 static void
 soup_socket_init (SoupSocket *sock)
@@ -103,18 +116,80 @@ soup_socket_init (SoupSocket *sock)
 	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
 	priv->non_blocking = TRUE;
+	priv->fd = -1;
 	g_mutex_init (&priv->addrlock);
 	g_mutex_init (&priv->iolock);
 }
 
+static gboolean
+soup_socket_initable_init (GInitable     *initable,
+			   GCancellable  *cancellable,
+			   GError       **error)
+{
+	SoupSocket *sock = SOUP_SOCKET (initable);
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	if (priv->conn) {
+		g_warn_if_fail (priv->gsock == NULL);
+		g_warn_if_fail (priv->fd == -1);
+
+		finish_socket_setup (sock);
+	}
+
+	if (priv->fd != -1) {
+		guint type, len = sizeof (type);
+
+		g_warn_if_fail (priv->gsock == NULL);
+
+		/* GSocket will g_error() this, so we have to check ourselves. */
+		if (getsockopt (priv->fd, SOL_SOCKET, SO_TYPE,
+				(gpointer)&type, (gpointer)&len) == -1) {
+			g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+					     _("Can't import non-socket as SoupSocket"));
+			return FALSE;
+		}
+
+		priv->gsock = g_socket_new_from_fd (priv->fd, error);
+		if (!priv->gsock)
+			return FALSE;
+	}
+
+	if (priv->gsock != NULL) {
+		int listening;
+
+		g_warn_if_fail (priv->local_addr == NULL);
+		g_warn_if_fail (priv->remote_addr == NULL);
+
+		if (!g_socket_get_option (priv->gsock,
+					  SOL_SOCKET, SO_ACCEPTCONN,
+					  &listening, error)) {
+			g_prefix_error (error, _("Could not import existing socket: "));
+			return FALSE;
+		}
+
+		finish_socket_setup (sock);
+		if (listening)
+			finish_listener_setup (sock);
+		else if (!g_socket_is_connected (priv->gsock)) {
+			g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_FAILED,
+					     _("Can't import unconnected socket"));
+			return FALSE;
+		}
+	}
+
+	return TRUE;
+}
+
 static void
 disconnect_internal (SoupSocket *sock, gboolean close)
 {
 	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 
 	g_clear_object (&priv->gsock);
-	if (priv->conn && close)
+	if (priv->conn && close) {
 		g_io_stream_close (priv->conn, NULL, NULL);
+		g_clear_object (&priv->conn);
+	}
 
 	if (priv->read_src) {
 		g_source_destroy (priv->read_src);
@@ -136,7 +211,7 @@ soup_socket_finalize (GObject *object)
 			g_warning ("Disposing socket %p during connect", object);
 		g_object_unref (priv->connect_cancel);
 	}
-	if (priv->gsock) {
+	if (priv->conn) {
 		if (priv->clean_dispose)
 			g_warning ("Disposing socket %p while still connected", object);
 		disconnect_internal (SOUP_SOCKET (object), TRUE);
@@ -150,7 +225,9 @@ soup_socket_finalize (GObject *object)
 	g_clear_object (&priv->local_addr);
 	g_clear_object (&priv->remote_addr);
 
+	g_clear_object (&priv->tls_interaction);
 	g_clear_object (&priv->proxy_resolver);
+	g_clear_object (&priv->ssl_creds);
 
 	if (priv->watch_src) {
 		if (priv->clean_dispose && !priv->is_server)
@@ -165,24 +242,28 @@ soup_socket_finalize (GObject *object)
 	G_OBJECT_CLASS (soup_socket_parent_class)->finalize (object);
 }
 
-
 static void
-finish_socket_setup (SoupSocketPrivate *priv)
+finish_socket_setup (SoupSocket *sock)
 {
-	if (!priv->gsock)
-		return;
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	if (priv->gsock) {
+		if (!priv->conn)
+			priv->conn = (GIOStream *)g_socket_connection_factory_create_connection (priv->gsock);
+
+		g_socket_set_timeout (priv->gsock, priv->timeout);
+		g_socket_set_option (priv->gsock, IPPROTO_TCP, TCP_NODELAY, TRUE, NULL);
+	}
 
 	if (!priv->conn)
-		priv->conn = (GIOStream *)g_socket_connection_factory_create_connection (priv->gsock);
+		return;
+
 	if (!priv->iostream)
 		priv->iostream = soup_io_stream_new (priv->conn, FALSE);
 	if (!priv->istream)
 		priv->istream = g_object_ref (g_io_stream_get_input_stream (priv->iostream));
 	if (!priv->ostream)
 		priv->ostream = g_object_ref (g_io_stream_get_output_stream (priv->iostream));
-
-	g_socket_set_timeout (priv->gsock, priv->timeout);
-	g_socket_set_option (priv->gsock, IPPROTO_TCP, TCP_NODELAY, TRUE, NULL);
 }
 
 static void
@@ -190,19 +271,34 @@ soup_socket_set_property (GObject *object, guint prop_id,
 			  const GValue *value, GParamSpec *pspec)
 {
 	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (object);
+	SoupSocketProperties *props;
 
 	switch (prop_id) {
+	case PROP_FD:
+		priv->fd = g_value_get_int (value);
+		break;
+	case PROP_GSOCKET:
+		priv->gsock = g_value_dup_object (value);
+		break;
+	case PROP_IOSTREAM:
+		priv->conn = g_value_dup_object (value);
+		break;
 	case PROP_LOCAL_ADDRESS:
-		priv->local_addr = (SoupAddress *)g_value_dup_object (value);
+		priv->local_addr = g_value_dup_object (value);
 		break;
 	case PROP_REMOTE_ADDRESS:
-		priv->remote_addr = (SoupAddress *)g_value_dup_object (value);
+		priv->remote_addr = g_value_dup_object (value);
 		break;
 	case PROP_NON_BLOCKING:
 		priv->non_blocking = g_value_get_boolean (value);
 		break;
+	case PROP_IPV6_ONLY:
+		priv->ipv6_only = g_value_get_boolean (value);
+		break;
 	case PROP_SSL_CREDENTIALS:
 		priv->ssl_creds = g_value_get_pointer (value);
+		if (priv->ssl_creds)
+			g_object_ref (priv->ssl_creds);
 		break;
 	case PROP_SSL_STRICT:
 		priv->ssl_strict = g_value_get_boolean (value);
@@ -211,23 +307,59 @@ soup_socket_set_property (GObject *object, guint prop_id,
 		priv->ssl_fallback = g_value_get_boolean (value);
 		break;
 	case PROP_ASYNC_CONTEXT:
-		priv->async_context = g_value_get_pointer (value);
-		if (priv->async_context)
-			g_main_context_ref (priv->async_context);
+		if (!priv->use_thread_context) {
+			priv->async_context = g_value_get_pointer (value);
+			if (priv->async_context)
+				g_main_context_ref (priv->async_context);
+		}
 		break;
 	case PROP_USE_THREAD_CONTEXT:
 		priv->use_thread_context = g_value_get_boolean (value);
+		if (priv->use_thread_context) {
+			g_clear_pointer (&priv->async_context, g_main_context_unref);
+			priv->async_context = g_main_context_ref_thread_default ();
+		}
 		break;
 	case PROP_TIMEOUT:
 		priv->timeout = g_value_get_uint (value);
 		if (priv->conn)
 			g_socket_set_timeout (priv->gsock, priv->timeout);
 		break;
-	case PROP_PROXY_RESOLVER:
-		priv->proxy_resolver = g_value_dup_object (value);
-		break;
-	case PROP_CLEAN_DISPOSE:
-		priv->clean_dispose = g_value_get_boolean (value);
+	case PROP_SOCKET_PROPERTIES:
+		props = g_value_get_boxed (value);
+		if (props) {
+			g_clear_pointer (&priv->async_context, g_main_context_unref);
+			if (props->use_thread_context) {
+				priv->use_thread_context = TRUE;
+				priv->async_context = g_main_context_ref_thread_default ();
+			} else {
+				priv->use_thread_context = FALSE;
+				if (props->async_context)
+					priv->async_context = g_main_context_ref (props->async_context);
+			}
+
+			g_clear_object (&priv->proxy_resolver);
+			if (props->proxy_resolver)
+				priv->proxy_resolver = g_object_ref (props->proxy_resolver);
+
+			g_clear_object (&priv->local_addr);
+			if (props->local_addr)
+				priv->local_addr = g_object_ref (props->local_addr);
+
+			g_clear_object (&priv->ssl_creds);
+			if (props->tlsdb)
+				priv->ssl_creds = g_object_ref (props->tlsdb);
+			g_clear_object (&priv->tls_interaction);
+			if (props->tls_interaction)
+				priv->tls_interaction = g_object_ref (props->tls_interaction);
+			priv->ssl_strict = props->ssl_strict;
+
+			priv->timeout = props->io_timeout;
+			if (priv->conn)
+				g_socket_set_timeout (priv->gsock, priv->timeout);
+
+			priv->clean_dispose = TRUE;
+		}
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -242,6 +374,9 @@ soup_socket_get_property (GObject *object, guint prop_id,
 	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (object);
 
 	switch (prop_id) {
+	case PROP_FD:
+		g_value_set_int (value, priv->fd);
+		break;
 	case PROP_LOCAL_ADDRESS:
 		g_value_set_object (value, soup_socket_get_local_address (SOUP_SOCKET (object)));
 		break;
@@ -251,6 +386,9 @@ soup_socket_get_property (GObject *object, guint prop_id,
 	case PROP_NON_BLOCKING:
 		g_value_set_boolean (value, priv->non_blocking);
 		break;
+	case PROP_IPV6_ONLY:
+		g_value_set_boolean (value, priv->ipv6_only);
+		break;
 	case PROP_IS_SERVER:
 		g_value_set_boolean (value, priv->is_server);
 		break;
@@ -284,8 +422,6 @@ soup_socket_get_property (GObject *object, guint prop_id,
 	case PROP_TLS_ERRORS:
 		g_value_set_flags (value, priv->tls_errors);
 		break;
-	case PROP_PROXY_RESOLVER:
-		g_value_set_object (value, priv->proxy_resolver);
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -400,6 +536,28 @@ soup_socket_class_init (SoupSocketClass *socket_class)
 
 
 	/* properties */
+	g_object_class_install_property (
+		 object_class, PROP_FD,
+		 g_param_spec_int (SOUP_SOCKET_FD,
+				   "FD",
+				   "The socket's file descriptor",
+				   -1, G_MAXINT, -1,
+				   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	g_object_class_install_property (
+		 object_class, PROP_GSOCKET,
+		 g_param_spec_object (SOUP_SOCKET_GSOCKET,
+				      "GSocket",
+				      "The socket's underlying GSocket",
+				      G_TYPE_SOCKET,
+				      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
+	g_object_class_install_property (
+		 object_class, PROP_IOSTREAM,
+		 g_param_spec_object (SOUP_SOCKET_IOSTREAM,
+				      "GIOStream",
+				      "The socket's underlying GIOStream",
+				      G_TYPE_IO_STREAM,
+				      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
+
 	/**
 	 * SOUP_SOCKET_LOCAL_ADDRESS:
 	 *
@@ -463,11 +621,28 @@ soup_socket_class_init (SoupSocketClass *socket_class)
 				      "Whether or not the socket uses non-blocking I/O",
 				      TRUE,
 				      G_PARAM_READWRITE));
+	g_object_class_install_property (
+		object_class, PROP_IPV6_ONLY,
+		g_param_spec_boolean (SOUP_SOCKET_IPV6_ONLY,
+				      "IPv6 only",
+				      "IPv6 only",
+				      FALSE,
+				      G_PARAM_READWRITE));
 	/**
 	 * SOUP_SOCKET_IS_SERVER:
 	 *
-	 * Alias for the #SoupSocket:is-server property. (Whether or
-	 * not the socket is a server socket.)
+	 * Alias for the #SoupSocket:is-server property, qv.
+	 **/
+	/**
+	 * SoupSocket:is-server:
+	 *
+	 * Whether or not the socket is a server socket.
+	 *
+	 * Note that for "ordinary" #SoupSockets this will be set for
+	 * both listening sockets and the sockets emitted by
+	 * #SoupSocket::new-connection, but for sockets created by
+	 * setting #SoupSocket:fd, it will only be set for listening
+	 * sockets.
 	 **/
 	g_object_class_install_property (
 		object_class, PROP_IS_SERVER,
@@ -579,13 +754,6 @@ soup_socket_class_init (SoupSocketClass *socket_class)
 				   0, G_MAXUINT, 0,
 				   G_PARAM_READWRITE));
 
-	g_object_class_install_property (
-		object_class, PROP_CLEAN_DISPOSE,
-		g_param_spec_boolean ("clean-dispose",
-				      "Clean dispose",
-				      "Warn on unclean dispose",
-				      FALSE,
-				      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
 	/**
 	 * SOUP_SOCKET_TLS_CERTIFICATE:
 	 *
@@ -622,12 +790,18 @@ soup_socket_class_init (SoupSocketClass *socket_class)
 				    G_PARAM_READABLE));
 
 	g_object_class_install_property (
-		object_class, PROP_PROXY_RESOLVER,
-		g_param_spec_object (SOUP_SOCKET_PROXY_RESOLVER,
-				     "Proxy resolver",
-				     "GProxyResolver to use",
-				     G_TYPE_PROXY_RESOLVER,
-				     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+		 object_class, PROP_SOCKET_PROPERTIES,
+		 g_param_spec_boxed (SOUP_SOCKET_SOCKET_PROPERTIES,
+				     "Socket properties",
+				     "Socket properties",
+				     SOUP_TYPE_SOCKET_PROPERTIES,
+				     G_PARAM_WRITABLE));
+}
+
+static void
+soup_socket_initable_interface_init (GInitableIface *initable_interface)
+{
+	initable_interface->init = soup_socket_initable_init;
 }
 
 
@@ -654,6 +828,15 @@ soup_socket_new (const char *optname1, ...)
 	return sock;
 }
 
+static void
+soup_socket_event (SoupSocket         *sock,
+		   GSocketClientEvent  event,
+		   GIOStream          *connection)
+{
+	g_signal_emit (sock, signals[EVENT], 0,
+		       event, connection);
+}
+
 static void
 re_emit_socket_client_event (GSocketClient       *client,
 			     GSocketClientEvent   event,
@@ -663,8 +846,7 @@ re_emit_socket_client_event (GSocketClient       *client,
 {
 	SoupSocket *sock = user_data;
 
-	g_signal_emit (sock, signals[EVENT], 0,
-		       event, connection);
+	soup_socket_event (sock, event, connection);
 }
 
 static gboolean
@@ -677,7 +859,7 @@ socket_connect_finish (SoupSocket *sock, GSocketConnection *conn)
 	if (conn) {
 		priv->conn = (GIOStream *)conn;
 		priv->gsock = g_object_ref (g_socket_connection_get_socket (conn));
-		finish_socket_setup (priv);
+		finish_socket_setup (sock);
 		return TRUE;
 	} else
 		return FALSE;
@@ -799,14 +981,10 @@ legacy_connect_async_cb (GObject       *object,
 			 gpointer       user_data)
 {
 	SoupSocket *sock = SOUP_SOCKET (object);
-	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
 	SoupSocketAsyncConnectData *sacd = user_data;
 	GError *error = NULL;
 	guint status;
 
-	if (priv->async_context && !priv->use_thread_context)
-		g_main_context_pop_thread_default (priv->async_context);
-
 	if (soup_socket_connect_finish_internal (sock, result, &error))
 		status = SOUP_STATUS_OK;
 	else
@@ -856,6 +1034,9 @@ soup_socket_connect_async (SoupSocket *sock, GCancellable *cancellable,
 	soup_socket_connect_async_internal (sock, cancellable,
 					    legacy_connect_async_cb,
 					    sacd);
+
+	if (priv->async_context && !priv->use_thread_context)
+		g_main_context_pop_thread_default (priv->async_context);
 }
 
 gboolean
@@ -942,6 +1123,34 @@ soup_socket_get_gsocket (SoupSocket *sock)
 	return SOUP_SOCKET_GET_PRIVATE (sock)->gsock;
 }
 
+GSocket *
+soup_socket_steal_gsocket (SoupSocket *sock)
+{
+	SoupSocketPrivate *priv;
+	GSocket *gsock;
+
+	g_return_val_if_fail (SOUP_IS_SOCKET (sock), NULL);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	gsock = priv->gsock;
+	priv->gsock = NULL;
+	g_clear_object (&priv->conn);
+	g_clear_object (&priv->iostream);
+
+	return gsock;
+}
+
+gboolean
+soup_socket_is_readable (SoupSocket *sock)
+{
+	SoupSocketPrivate *priv;
+
+	g_return_val_if_fail (SOUP_IS_SOCKET (sock), FALSE);
+	priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	return g_pollable_input_stream_is_readable (G_POLLABLE_INPUT_STREAM (priv->istream));
+}
+
 GIOStream *
 soup_socket_get_connection (SoupSocket *sock)
 {
@@ -964,7 +1173,6 @@ soup_socket_create_watch (SoupSocketPrivate *priv, GIOCondition cond,
 			  GCancellable *cancellable)
 {
 	GSource *watch;
-	GMainContext *async_context;
 
 	if (cond == G_IO_IN)
 		watch = g_pollable_input_stream_create_source (G_POLLABLE_INPUT_STREAM (priv->istream), cancellable);
@@ -972,12 +1180,7 @@ soup_socket_create_watch (SoupSocketPrivate *priv, GIOCondition cond,
 		watch = g_pollable_output_stream_create_source (G_POLLABLE_OUTPUT_STREAM (priv->ostream), cancellable);
 	g_source_set_callback (watch, (GSourceFunc)callback, user_data, NULL);
 
-	if (priv->use_thread_context)
-		async_context = g_main_context_get_thread_default ();
-	else
-		async_context = priv->async_context;
-
-	g_source_attach (watch, async_context);
+	g_source_attach (watch, priv->async_context);
 	g_source_unref (watch);
 
 	return watch;
@@ -1001,11 +1204,12 @@ listen_watch (GObject *pollable, gpointer data)
 		new_priv->async_context = g_main_context_ref (priv->async_context);
 	new_priv->use_thread_context = priv->use_thread_context;
 	new_priv->non_blocking = priv->non_blocking;
+	new_priv->clean_dispose = priv->clean_dispose;
 	new_priv->is_server = TRUE;
 	new_priv->ssl = priv->ssl;
 	if (priv->ssl_creds)
-		new_priv->ssl_creds = priv->ssl_creds;
-	finish_socket_setup (new_priv);
+		new_priv->ssl_creds = g_object_ref (priv->ssl_creds);
+	finish_socket_setup (new);
 
 	if (new_priv->ssl_creds) {
 		if (!soup_socket_start_proxy_ssl (new, NULL, NULL)) {
@@ -1020,6 +1224,17 @@ listen_watch (GObject *pollable, gpointer data)
 	return TRUE;
 }
 
+static void
+finish_listener_setup (SoupSocket *sock)
+{
+	SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+	priv->is_server = TRUE;
+	priv->watch_src = soup_socket_create_watch (priv, G_IO_IN,
+						    listen_watch, sock,
+						    NULL);
+}
+
 /**
  * soup_socket_listen:
  * @sock: a server #SoupSocket (which must not already be connected or
@@ -1042,8 +1257,6 @@ soup_socket_listen (SoupSocket *sock)
 	g_return_val_if_fail (priv->gsock == NULL, FALSE);
 	g_return_val_if_fail (priv->local_addr != NULL, FALSE);
 
-	priv->is_server = TRUE;
-
 	/* @local_addr may have its port set to 0. So we intentionally
 	 * don't store it in priv->local_addr, so that if the
 	 * caller calls soup_socket_get_local_address() later, we'll
@@ -1059,7 +1272,18 @@ soup_socket_listen (SoupSocket *sock)
 				    NULL);
 	if (!priv->gsock)
 		goto cant_listen;
-	finish_socket_setup (priv);
+	finish_socket_setup (sock);
+
+#if defined (IPPROTO_IPV6) && defined (IPV6_V6ONLY)
+	if (priv->ipv6_only) {
+		int fd, v6_only;
+
+		fd = g_socket_get_fd (priv->gsock);
+		v6_only = TRUE;
+		setsockopt (fd, IPPROTO_IPV6, IPV6_V6ONLY,
+			    &v6_only, sizeof (v6_only));
+	}
+#endif
 
 	/* Bind */
 	if (!g_socket_bind (priv->gsock, addr, TRUE, NULL))
@@ -1071,10 +1295,8 @@ soup_socket_listen (SoupSocket *sock)
 	/* Listen */
 	if (!g_socket_listen (priv->gsock, NULL))
 		goto cant_listen;
+	finish_listener_setup (sock);
 
-	priv->watch_src = soup_socket_create_watch (priv, G_IO_IN,
-						    listen_watch, sock,
-						    NULL);
 	g_object_unref (addr);
 	return TRUE;
 
@@ -1140,6 +1362,14 @@ soup_socket_setup_ssl (SoupSocket    *sock,
 		if (!conn)
 			return FALSE;
 
+		/* GLib < 2.41 mistakenly doesn't implement this property in the
+		 * dummy TLS backend, so we don't include it in the g_initable_new()
+		 * call above.
+		 */
+		g_object_set (G_OBJECT (conn),
+			      "interaction", priv->tls_interaction,
+			      NULL);
+
 		g_object_unref (priv->conn);
 		priv->conn = G_IO_STREAM (conn);
 
@@ -1195,7 +1425,7 @@ soup_socket_start_ssl (SoupSocket *sock, GCancellable *cancellable)
 	return soup_socket_setup_ssl (sock, soup_address_get_name (priv->remote_addr),
 				      cancellable, NULL);
 }
-	
+
 /**
  * soup_socket_start_proxy_ssl:
  * @sock: the socket
@@ -1225,8 +1455,14 @@ soup_socket_handshake_sync (SoupSocket    *sock,
 	if (!soup_socket_setup_ssl (sock, ssl_host, cancellable, error))
 		return FALSE;
 
-	return g_tls_connection_handshake (G_TLS_CONNECTION (priv->conn),
-					   cancellable, error);
+	soup_socket_event (sock, G_SOCKET_CLIENT_TLS_HANDSHAKING, priv->conn);
+
+	if (!g_tls_connection_handshake (G_TLS_CONNECTION (priv->conn),
+					 cancellable, error))
+		return FALSE;
+
+	soup_socket_event (sock, G_SOCKET_CLIENT_TLS_HANDSHAKED, priv->conn);
+	return TRUE;
 }
 
 static void
@@ -1236,9 +1472,13 @@ handshake_async_ready (GObject *source, GAsyncResult *result, gpointer user_data
 	GError *error = NULL;
 
 	if (g_tls_connection_handshake_finish (G_TLS_CONNECTION (source),
-					       result, &error))
+					       result, &error)) {
+		SoupSocket *sock = g_task_get_source_object (task);
+		SoupSocketPrivate *priv = SOUP_SOCKET_GET_PRIVATE (sock);
+
+		soup_socket_event (sock, G_SOCKET_CLIENT_TLS_HANDSHAKED, priv->conn);
 		g_task_return_boolean (task, TRUE);
-	else
+	} else
 		g_task_return_error (task, error);
 	g_object_unref (task);
 }
@@ -1262,6 +1502,8 @@ soup_socket_handshake_async (SoupSocket          *sock,
 		return;
 	}
 
+	soup_socket_event (sock, G_SOCKET_CLIENT_TLS_HANDSHAKING, priv->conn);
+
 	g_tls_connection_handshake_async (G_TLS_CONNECTION (priv->conn),
 					  G_PRIORITY_DEFAULT,
 					  cancellable, handshake_async_ready,
@@ -1313,7 +1555,7 @@ soup_socket_disconnect (SoupSocket *sock)
 		g_cancellable_cancel (priv->connect_cancel);
 		return;
 	} else if (g_mutex_trylock (&priv->iolock)) {
-		if (priv->gsock)
+		if (priv->conn)
 			disconnect_internal (sock, TRUE);
 		else
 			already_disconnected = TRUE;
@@ -1544,7 +1786,8 @@ translate_read_status (SoupSocket *sock, GCancellable *cancellable,
 /**
  * soup_socket_read:
  * @sock: the socket
- * @buffer: buffer to read into
+ * @buffer: (array length=len) (element-type guint8): buffer to read
+ *   into
  * @len: size of @buffer in bytes
  * @nread: (out): on return, the number of bytes read into @buffer
  * @cancellable: a #GCancellable, or %NULL
@@ -1609,7 +1852,8 @@ out:
 /**
  * soup_socket_read_until:
  * @sock: the socket
- * @buffer: buffer to read into
+ * @buffer: (array length=len) (element-type guint8): buffer to read
+ *   into
  * @len: size of @buffer in bytes
  * @boundary: boundary to read until
  * @boundary_len: length of @boundary in bytes
@@ -1685,7 +1929,7 @@ socket_write_watch (GObject *pollable, gpointer user_data)
 /**
  * soup_socket_write:
  * @sock: the socket
- * @buffer: data to write
+ * @buffer: (array length=len) (element-type guint8): data to write
  * @len: size of @buffer, in bytes
  * @nwrote: (out): on return, number of bytes written
  * @cancellable: a #GCancellable, or %NULL
